# -*- coding: utf-8 -*-
#
# This file is part of EventGhost.
# Copyright © 2005-2016 EventGhost Project <http://www.eventghost.net/>
#
# EventGhost is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option)
# any later version.
#
# EventGhost is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with EventGhost. If not, see <http://www.gnu.org/licenses/>.

import re
import os
import socket
import requests
import sys
import xml.etree.cElementTree as ElementTree
from urlparse import urlparse
from shutil import copyfile


HEADER = """# -*- coding: utf-8 -*-
#
# This file is part of EventGhost.
# Copyright © 2005-2016 EventGhost Project <http://www.eventghost.net/>
#
# EventGhost is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option)
# any later version.
#
# EventGhost is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with EventGhost. If not, see <http://www.gnu.org/licenses/>.
#
#
# ******************* THIS FILE IS AUTOMATICALLY GENERATED *******************
# ******************************* DO NOT MODIFY ******************************

"""


'http://192.168.1.2/cgi-bin/cmh/view_upnp_file.sh?file=S_HomeAutomationGateway1.xml.lzo'
URL = 'http://{ip_address}/cgi-bin/cmh/'
GET_UPNP_FILES = URL + 'get_upnp_files.sh'
VIEW_UPNP_FILE = URL + 'view_upnp_file.sh?file={file}'
SYS_INFO = URL + 'sysinfo.sh'
CATEGORIES = 'http://{ip_address}/cmh/js/config/constants.js'
VERA_INFO = 'http://{ip_address}/upnp/vera.xml'

CORE_FILE_PATH = os.path.join(os.path.split(__file__)[0], 'core')

if os.name == 'nt':
    BUILD_PATH = os.path.join(
        os.path.expandvars('%APPDATA%'),
        'MiCasaVerde_Vera'
    )

else:
    BUILD_PATH = os.path.join(os.path.expanduser('~'), '.MiCasaVerde_Vera')

if not os.path.exists(BUILD_PATH):
    os.mkdir(BUILD_PATH)
    # for core_file in os.listdir(CORE_FILE_PATH):

    #    copyfile(
    #        os.path.join(CORE_FILE_PATH, core_file),
    #        os.path.join(BUILD_PATH, core_file)
    #    )

# sys.path.append(BUILD_PATH)

SSDP_ADDR = "239.255.255.250"
SSDP_PORT = 1900
SSDP_MX = 10
SSDP_ST = "upnp:rootdevice"

SSDP_REQUEST = (
    'M-SEARCH * HTTP/1.1\r\n'
    'MX: %d\r\n'
    'ST: %s\r\n'
    'HOST: %s:%d\r\n'
    'MAN: "ssdp:discover"\r\n'
    '\r\n' % (SSDP_MX, SSDP_ST, SSDP_ADDR, SSDP_PORT)
)


LOCATION_MASK = re.compile("location:[ ]*(.+)\r\n", re.IGNORECASE)


DEVICE_SUBCLASS_IMPORT = '''from {module_name} import {class_name}\n'''
DEVICE_SUBCLASS_INIT_TEMPLATE = '''        {class_name}.__init__(self, parent)\n'''
DEVICE_CLASS_TEMPLATE = '''
{imports}
class {class_name}({subclasses}):
    _service_id = '{device_type}'
    _variables = dict()
    
    def __init__(self, parent, node):
        self._parent = parent
{subclass_inits}
        self._variables.update(node)
    
    def __getattr__(self, item):
        if item in self.__dict__:
            return self.__dict__[item]
        if item in self._variables:
            attr = self._variables[item]
            if attr is None:
                raise AttributeError
            return attr
        if item.replace('.', '') in self._variables:
            attr = self._variables[item.replace('.', '')]
            if attr is None:
                raise AttributeError
            return attr
            
        raise AttributeError
        
    def __setattr__(self, key, value):
        if key.startswith('_'):
            object.setattr(self, key, value)
        else:
            raise AttributeError
            
    def node(self, node):
        for key, value in node.items():
            if key in self._variables and self._variables[key] != value:
                self._variables[key] = value
                self._parent.trigger_event(
                    device=self,
                    variable=key,
                    value=value
                )
    
    node = property(fset=node)
'''


CLASS_TEMPLATE = '''
_default_variables = {{{attributes}
}}


class {class_name}(object):
    """
    Attributes:
{class_doc}
    """
    
    _variables = dict()
    
    def __init__(self, parent):
        self._parent = parent
        defaults = _default_variables.copy()
        self._variables.update(defaults)
{methods}

'''

ATTR_DOC_TEMPLATE = """        {attr_name} {attr_type}: {attr_docs}\n"""
STATE_TEMPLATE = "\n    '{attr_name}': None,"
ALLOWED_VALUES_TEMPLATE = '''
            Allowed Values:
                {values}
'''

ALLOWED_RANGE_TEMPLATE = '''
            Allowed Range:
                {values}
'''

DATA_TYPES = dict(
    string='(str)',
    char='(str)',
    boolean='(bool)',
    ui4='(int)',
    ui1='(int)',
    ui2='(int)',
    i1='(int)',
    number='(int)',
    i4='(int)',
    float='(float)'
)

SEND_TEMPLATE = '''        self._parent.send({send_arguments}
        )'''
SEND_ARGUMENT_TEMPLATE = '''
            {keyword}={value},'''

PROPERTY_TEMPLATE = '''
    @property
    def {method}(self):
        return self._variables['{method}']

    @{method}.setter
    def {method}(self, value):
{send_template}
'''

METHOD_TEMPLATE = '''
    def {method}(self, {keywords}):
{send_template}
'''


CLASS_ITEMS = dict()
vera_files = os.listdir(BUILD_PATH)


def create_service_name(service_type):
    service_type = service_type.split(':')

    if len(service_type) == 5:
        return ''.join(service_type[-2:])
    else:
        return service_type[-1]


def parse_string(word):
    next_to_last_char = ''
    last_char = ''
    new_word = ''
    for char in word:
        if last_char and (char.isupper() or char.isdigit()):
            if last_char == '_':
                new_word = new_word[:-1]
            elif last_char.isupper():
                new_word = new_word[:-2] + last_char.lower()
            new_word += '_'

        elif char == '_' and last_char.isupper():
            new_word = new_word[:-2] + last_char.lower()
        new_word += char.lower()
        last_char = char
    if last_char.isupper():
        new_word = new_word[:-2] + last_char.lower()
    return new_word


def discover():
    dest = socket.gethostbyname(SSDP_ADDR)

    def connect_sock(bind=False):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        if bind:
            sock.bind((socket.gethostbyname(socket.gethostname()), 0))
        sock.sendto(SSDP_REQUEST, (dest, SSDP_PORT))
        sock.settimeout(7)

        try:
            while True:
                data, address = sock.recvfrom(1024) # buffer size is 1024 bytes
                if 'luaupnp.xml' in data:
                    return address[0]
        except socket.timeout:
            sock.close()

    ip_address = connect_sock()
    if ip_address is None:
        ip_address = connect_sock(bind=True)

    return ip_address


def get_categories(ip_address):
    response = requests.get(
        CATEGORIES.format(ip_address=ip_address),
        timeout=5
    )
    data = response.content

    data = data[data.find('CategoryNameById = {') +19:]
    categories = eval(data[:data.find('}')])
    data = data[data.find('KnownDeviceTypes = [') - 1:]
    known_types = data[:data.find('];')]

    for category_number, category_type in categories.items()[:]:
        data = known_types[known_types.find(category_type):]
        data = data[:data.find('},') + 1]
        category_name = data[data.find('text') + 7:data.rfind("'") - 1]
        categories[category_number.replace('_', '.')] = category_name


def get_vera_info(ip_address):
    response = requests.get(
        SYS_INFO.format(ip_address = ip_address),
        timeout=5
    )
    data = response.json()
    serial_number = data['installation_number']
    fw_version = data['firmware_version']
    zw_version = data['zwave_version']
    home_id = data['zwave_homeid']

    response = requests.get(
        VERA_INFO.format(ip_address = ip_address),
        timeout=5
    )

    xml_root = ElementTree.fromstring(response.content)

    xml_root = xml_root.find(
        '{urn:schemas-upnp-org:device-1-0}'
        'device'
    )
    manufacturer = xml_root.find(
        '{urn:schemas-upnp-org:device-1-0}'
        'manufacturer'
    ).text
    model_name = xml_root.find(
        '{urn:schemas-upnp-org:device-1-0}'
        'modelName'
    ).text
    model_number = xml_root.find(
        '{urn:schemas-upnp-org:device-1-0}'
        'modelNumber'
    ).text
    model_description = xml_root.find(
        '{urn:schemas-upnp-org:device-1-0}'
        'modelDescription'
    ).text

    return (
        manufacturer,
        model_name,
        model_number,
        model_description,
        serial_number,
        fw_version,
        zw_version,
        home_id
    )


def parse_location(ip_address):

    found_services = invoke(ip_address)

    response = requests.get(
        GET_UPNP_FILES.format(ip_address=ip_address),
        timeout=5
    )
    device_files = ['vera.xml.lzo'] + list(
        file_name for file_name in response.content.split('\n')
        if file_name.startswith('D_') and file_name.endswith('.xml.lzo')
    )

    for device_file in device_files:
        response = requests.get(
            VIEW_UPNP_FILE.format(
                ip_address=ip_address,
                file=device_file
            ),
            timeout=5
        )
        response = response.content
        if 'xmlns' in response:
            root_xmlns = (
                '{%s}' %
                response[response.find('xmlns="') + 7:response.find('">')]
            )
        else:
            root_xmlns = ''
        try:
            xml_root = ElementTree.fromstring(response)
        except:
            print('\t[!] Failed XML parsing of %s' % device_file)
        else:
            device_type = xml_root.find('%sdeviceType' % root_xmlns)

            if device_type in found_services:
                continue

            subclasses = []

            services = xml_root.findall('.//*%sserviceList/' % root_xmlns)

            for service in services:
                service_type = service.find('%sserviceType' % root_xmlns).text

                if service_type in found_services:
                    svc = found_services[service_type]
                    methods = svc['methods']
                    attributes = svc['attributes']
                    class_doc = svc['class_doc']
                    properties = svc['properties']
                else:
                    methods = dict()
                    attributes = []
                    class_doc = ''
                    properties = dict()

                service_id = service.find('%sserviceId' % root_xmlns).text

                print('-Processing ' + service_type + '.....')

                subclasses += [service_type]

                scpd_url = service.find('%sSCPDURL' % root_xmlns).text

                response = requests.get(
                    VIEW_UPNP_FILE.format(
                        ip_address=ip_address,
                        file=scpd_url
                    ) + '.lzo',
                    timeout=5
                )

                response = response.content
                if 'xmlns' in response:
                    service_xmlns = (
                        '{%s}' %
                        response[response.find('xmlns="') + 7:response.find('">')]
                    )
                else:
                    service_xmlns = ''

                try:
                    service_xml = ElementTree.fromstring(response)
                except SyntaxError:
                    print response
                    continue

                actions = service_xml.find('%sactionList' % service_xmlns)

                if actions is not None:
                    create_class_methods(
                        service_xmlns,
                        service_id,
                        actions,
                        methods,
                        attributes
                    )

                state_variables = service_xml.find(
                    '%sserviceStateTable' % service_xmlns
                )
                if state_variables is not None:
                    docs = create_class_attributes(
                        service_xmlns,
                        state_variables,
                        attributes
                    )
                    class_doc += docs

                found_services[service_type] = dict(
                    methods=methods,
                    attributes=attributes,
                    class_doc=class_doc,
                    properties=properties
                )

            if device_type is not None:
                found_services[device_type] = dict(subclasses=subclasses)

    build_templates(found_services)


def build_templates(services):
    print_dict(services)

    for service_type, params in services.items():

        service_name = create_service_name(service_type)
        class_name = service_name.replace('_', '')
        file_name = parse_string(class_name) + '.py'

        if 'subclasses' in params:
            imports = []
            subclasses = []
            subclass_inits = []

            for subclass in params['subclasses']:
                subclass_name = create_service_name(subclass).replace('_', '')
                subclass_module = parse_string(subclass_name)
                subclasses += [subclass_name]

                imports += [
                    DEVICE_SUBCLASS_IMPORT.format(
                        module=subclass_module,
                        class_name=subclass_name
                    )
                ]
                subclass_inits += [
                    DEVICE_SUBCLASS_INIT_TEMPLATE.format(
                        class_name=subclass_name
                    )
                ]

            template = DEVICE_CLASS_TEMPLATE.format(
                device_type=service_type,
                class_name=class_name,
                subclasses=', '.join(subclasses),
                subclass_inits=''.join(subclass_inits),
                imports=''.join(imports)
            )
        else:
            template = make_class_template(service_type, **params)

        template = template.replace('self, ):', 'self):').replace(',\n        )', '\n        )')

        file_path = os.path.join(BUILD_PATH, file_name)

        print('Writing File {0} ....'.format(file_path))

        with open(file_path, 'w') as f:
            f.write(HEADER)
            f.write(template)


def create_class_methods(
    service_xmlns,
    service_id,
    actions,
    methods,
    attributes
):
    for action in actions:
        name = action.find('%sname' % service_xmlns).text
        arguments = action.find('%sargumentList' % service_xmlns)

        send_arguments = [
            ['id', "'action'"],
            ['serviceId', '%r' % service_id],
            ['action', '%r' % name]
        ]
        keywords = []

        if arguments is not None:
            for argument in arguments:
                direction = argument.find('%sdirection' % service_xmlns)
                related = argument.find(
                    '%srelatedStateVariable' % service_xmlns
                )

                if related is not None and related.text not in attributes:
                    attributes += [related.text]

                if direction is not None and direction.text == 'in':
                    arg_name = argument.find('%sname' % service_xmlns).text

                    if arg_name == 'DataFormat':
                        continue

                    p_arg_name = parse_string(arg_name)
                    if p_arg_name == 'reload':
                        p_arg_name = 'lua_reload'

                    send_arguments += [[arg_name, p_arg_name]]
                    keywords += [p_arg_name]

        p_name = parse_string(name)

        if p_name not in methods:
            methods[p_name] = [keywords, send_arguments]


def create_class_attributes(service_xmlns, state_variables, attributes):
    class_doc_templates = []

    for state_variable in state_variables:
        attr_name = state_variable.find('%sname' % service_xmlns).text

        data_type = state_variable.find('%sdataType' % service_xmlns).text,

        data_type = DATA_TYPES.get(
            data_type,
            data_type
        )

        attr_docs = ''
        allowed_values = state_variable.find(
            '%sallowedValueList' % service_xmlns
        )
        if allowed_values is not None:
            a_values = []
            for value in allowed_values:
                a_values += [value.text]

            attr_docs += ALLOWED_VALUES_TEMPLATE.format(
                values='\n                '.join(a_values)
            )

        allowed_range = state_variable.find(
            '%sallowedValueRange' % service_xmlns
        )
        if allowed_range is not None:
            a_values = []
            for value in allowed_range:
                a_values += ['%s: %s' % (
                    value.tag.replace(
                        service_xmlns,
                        ''
                    ),
                    value.text
                )]
            attr_docs += ALLOWED_RANGE_TEMPLATE.format(
                values='\n                '.join(a_values)
            )

        attr_name = attr_name.replace('.', '')

        if attr_name not in attributes:
            attributes += [attr_name]

        class_doc_templates += [
            ATTR_DOC_TEMPLATE.format(
                attr_name=attr_name,
                attr_type=data_type,
                attr_docs=attr_docs
            )
        ]

    return ''.join(class_doc_templates)


def make_property_template(method, send_arguments):
    send_template = SEND_TEMPLATE.format(
        send_arguments=send_arguments
    )
    template = PROPERTY_TEMPLATE.format(
        method=method,
        send_template=send_template
    )
    return template


def make_method_template(method, keywords, send_arguments):
    send_template = SEND_TEMPLATE.format(
        send_arguments=send_arguments
    )
    template = METHOD_TEMPLATE.format(
        keywords=', '.join(keywords),
        method=method,
        send_template=send_template
    )
    return template


def make_class_template(
    service_type,
    methods,
    attributes,
    class_doc,
    properties
):
    service_name = create_service_name(service_type)
    class_name = service_name.replace('_', '')

    cls_methods = ''
    for method_name, params in methods.items():
        keywords, send_arguments = params

        send_arguments = ''.join(
            SEND_ARGUMENT_TEMPLATE.format(keyword=keyword, value=value)
            for keyword, value in send_arguments
        )

        cls_methods += make_method_template(
            method_name,
            keywords,
            send_arguments
        )

    cls_properties = ''
    for property_name, send_arguments in properties.items():
        send_arguments = ''.join(
            SEND_ARGUMENT_TEMPLATE.format(keyword=keyword, value=value)
            for keyword, value in send_arguments
        )

        cls_properties += make_property_template(property_name, send_arguments)

    cls_attributes = ''.join(
        STATE_TEMPLATE.format(attr_name=attribute)
        for attribute in attributes
    )

    template = CLASS_TEMPLATE.format(
        class_name=class_name,
        methods=cls_methods + cls_properties,
        attributes=cls_attributes,
        class_doc=class_doc

    )

    return template


def invoke(ip_address):
    url = 'http://{ip}:3480/data_request'.format(ip=ip_address)

    response = requests.get(url, params=dict(id='invoke'))
    data = response.content
    services = dict()

    for line in data.split('\n'):
        found_device = line.find('<a href=')
        if found_device > -1:
            task = line[line.find('"> ') + 3:line.find('</a>')]
            line = (
                line[found_device + 9:line.find('">')].replace(
                    'lu_invoke',
                    'invoke'
                )
            )
            if 'RunScene' not in line:
                service_type = ''
                response = requests.get(
                    'http://{ip}:3480/{path}'.format(ip=ip_address, path=line)
                )
                device_data = response.content
                methods = dict()
                attributes = []
                properties = dict()
                class_doc = ''
                for device_line in device_data.split('\n'):
                    command_found = device_line.find('<a href=')
                    found_service = device_line.find('<br><i>')
                    if found_service > -1:
                        if service_type:
                            services[service_type] = dict(
                                methods=methods,
                                attributes=attributes,
                                class_doc=class_doc,
                                properties=properties
                            )

                            methods = dict()
                            attributes = []
                            class_doc = ''
                            properties = dict()

                        service_id = (
                            device_line[found_service + 7: device_line.find('</i>')]
                        )

                        service_type = ''
                        while service_id[-1].isdigit():
                            service_type = service_type + service_id[-1]
                            service_id = service_id[:-1]

                        service_type = service_id + ':' + service_type
                        service_id += service_type.split(':')[-1]

                        if service_type in services:
                            svc = services[service_type]
                            methods = svc['methods']
                            attributes = svc['attributes']
                            class_doc = svc['class_doc']
                            properties = svc['properties']
                        else:
                            print('-Processing ' + service_type + '.....')
                            services[service_type] = dict()

                    if command_found > -1:
                        send_arguments = []
                        method = None
                        keywords = []

                        command_line = device_line[command_found + 9:]
                        command = command_line[:command_line.find('">')]

                        command = command.replace(
                            'lu_variableset',
                            'variableset'
                        ).replace(
                            'data_request?',
                            ''
                        ).split('&')

                        for param in command:
                            keyword, value = param.split('=', 1)

                            if keyword == 'action':
                                method_name = parse_string(value)

                                if method_name in methods:
                                    continue

                                method = [
                                    make_method_template,
                                    method_name
                                ]

                            if keyword == 'Variable':
                                property_name = value.replace('.', '')

                                if property_name in properties:
                                    continue

                                method = [
                                    make_property_template,
                                    property_name
                                ]

                            if not value:
                                value = parse_string(keyword)
                                if value == 'reload':
                                    value = 'lua_reload'
                                keywords += [', ' + value]

                            elif keyword == 'Value':
                                value = 'value'
                                keywords += ['value']

                            elif keyword == 'DeviceNum':
                                value = 'self.DeviceNum'

                            else:
                                value = '%r' % value

                            send_arguments += [[keyword, value]]

                        if method:
                            template, method = method

                            if method == create_class_methods:
                                methods[method] = [keywords, send_arguments]

                            else:
                                properties[method] = send_arguments
                                if method not in attributes:
                                    attributes += [method]

                if service_type:
                    svc = services[service_type]
                    svc['methods'] = methods
                    svc['attributes'] = attributes
                    svc['class_doc'] = class_doc
                    svc['properties'] = properties
    return services


def print_list(l, indent):

    for item in l:
        if isinstance(item, list):
            if len(item) > 2:
                print indent + '['
                print_list(item, indent + '    ')
                print indent + ']'
            else:
                print indent + repr(item)
        else:
            print indent + repr(item)


def print_dict(d, indent=''):
    for key, value in d.items():
        if isinstance(value, dict):
            print indent, key, ':'
            print_dict(value, indent + '    ')
        elif isinstance(value, list):
            print indent + key + ':'
            print_list(value, indent + '    ')
        else:
            print indent + key + ':', value


def main(ip_address):
    if not ip_address:
        ip_address = discover()

    if ip_address is not None:
        (
            manufacturer,
            model_name,
            model_number,
            model_description,
            serial_number,
            fw_version,
            zw_version,
            home_id
        ) = get_vera_info(ip_address)
        print('-' * 13 + ' Vera Located ' + '-' * 13)
        print('')
        print('-Manufacturer: ' + manufacturer)
        print('-Model Name: ' + model_name)
        print('-Model Number: ' + model_number)
        print('-Model Description: ' + model_description)
        print('-Serial Number: ' + serial_number)
        print('-Firmware Version: ' + fw_version)
        print('-ZWave Version: ' + zw_version)
        print('-Home Id: ' + home_id)
        print('-' * 40)
        print('')

        parse_location(ip_address)

    return ip_address

if __name__ == "__main__":
    ip = raw_input('input ip address of vera\nor leave blank for auto discovery\n')
    main(ip)
